#include "stdafx.h"
#include "Skills.h"
#include "LogManager.h"
#include "Config.h"
#include "WorldObjects.h"

#define POWERRANGE(x)  (((x)>cfg->maxpower)?cfg->maxpower:             \
                        (((x)<cfg->minpower)?cfg->minpower:(x)))
#define MOMENTRANGE(x) (((x)>cfg->maxmoment)?cfg->maxmoment:           \
                        (((x)<cfg->minmoment)?cfg->minmoment:(x)))
#define NECKRANGE(x)   (((x)>cfg->maxneckmoment)?cfg->maxneckmoment:   \
                        (((x)<cfg->minneckmoment)?cfg->minneckmoment:(x)))

extern CConfig *cfg;
extern FILE *m_ff;

void CBasicSkills::Dash(CLocalAgent *pa, double power)
{
    CGameCommand *pGameCommand = new CGameCommand();

    pGameCommand->Dash(POWERRANGE(power));
    pa->actionseq->AddGameCommand(pGameCommand);
}

void CBasicSkills::Turn(CLocalAgent *pa, double moment)
{
    CGameCommand *pGameCommand = new CGameCommand();
    
    double moment2 = RadToDeg(moment);  moment2 = MOMENTRANGE(moment2);
    pGameCommand->Turn(moment2);
    pa->actionseq->AddGameCommand(pGameCommand);

//    // QUICK FIX NA OTACANIE - ZRUSIT KED BUDE DOBRY WORLDMODEL (SO ZAPOCITANIM POSLEDNEJ HRACOVEJ AKCIE)
//    pa->lastturn = DegToRad(moment2);
}

void CBasicSkills::Kick(CLocalAgent *pa, double power, double dir)
{
    CGameCommand *pGameCommand = new CGameCommand();
    
    pGameCommand->Kick(POWERRANGE(power), RadToDeg(dir));
    pa->actionseq->AddGameCommand(pGameCommand);
}

void CBasicSkills::TurnNeck(CLocalAgent *pa, double moment)
{
    CGameCommand *pGameCommand = new CGameCommand();
    
    double moment2 = RadToDeg(moment);  moment2 = NECKRANGE(moment2);
    pGameCommand->TurnNeck(RadToDeg(moment2));
    pa->actionseq->AddGameCommand(pGameCommand);
}

void CBasicSkills::Move(CLocalAgent *pa, double x, double y)
{
    CGameCommand *pGameCommand = new CGameCommand();
    
    pGameCommand->Move(x, y);
    pa->actionseq->AddGameCommand(pGameCommand);
}

void CBasicSkills::Move(CLocalAgent *pa, C2DVector pos)
{
    Move(pa, pos.GetX(), pos.GetY());
}

void CBasicSkills::Catch(CLocalAgent *pa, double dir)
{
    CGameCommand *pGameCommand = new CGameCommand();
    
    pGameCommand->Catch(RadToDeg(dir));
    pa->actionseq->AddGameCommand(pGameCommand);
}

void CBasicSkills::Say(CLocalAgent *pa, string message)
{
    CGameCommand *pGameCommand = new CGameCommand();
    
    pGameCommand->Say(message);
    pa->actionseq->AddGameCommand(pGameCommand);
}

void CBasicSkills::ChangeView(CLocalAgent *pa, EViewWidth width, EViewQuality quality)
{
    CGameCommand *pGameCommand = new CGameCommand();
    
    pGameCommand->ChangeView(width, quality);
    pa->actionseq->AddGameCommand(pGameCommand);
}



void CBasicSkills::Dash2(CLocalAgent *pa, double accel)
{
    Dash(pa, accel/(cfg->dash_power_rate));
}


void CBasicSkills::Turn2(CLocalAgent *pa, double angle)
{
    double moment = TurnReal2Wanted(pa, angle);
    if ((moment > DegToRad(180)) || (moment < DegToRad(-180)))
    {
        LOGMSG(1, "ERROR Skill 'Turn2': greater angle than possible (%lf) (speed %lf)", moment, pa->GetSpeed().GetDistance());
//        return;
        moment > DegToRad(180) ? moment = DegToRad(180) : moment = DegToRad(-180);
    }
    Turn(pa, moment);
}

double CBasicSkills::TurnWanted2Real(CLocalAgent *pa, double angle)
{
    double dPlayerSpeed = pa->GetSpeed().GetDistance();
//    return DegToRad(RadToDeg(angle) / (1.0 + cfg->inertia_moment * dPlayerSpeed));
    return angle / (1.0 + cfg->inertia_moment * dPlayerSpeed);
}

double CBasicSkills::TurnReal2Wanted(CLocalAgent *pa, double moment)
{
    double dPlayerSpeed = pa->GetSpeed().GetDistance();
//    return DegToRad(RadToDeg(angle) / (1.0 + cfg->inertia_moment * dPlayerSpeed));
    return moment * (1.0 + cfg->inertia_moment * dPlayerSpeed);
}



void CBasicSkills::Kick2(CLocalAgent *pa, double accel, double dir)
{
    // treba dat pozor aby sme nekopli do seba (prip. do ineho hraca???)

    // treba zapocitat rychlost lopty + natocenie
    CBall *pBall = pa->GetWorldModel()->GetBall();
    CDiffuse2DVector vball = Rot(pBall->GetPosition() - pa->GetPosition(), -pa->GetDirection());
    double dist_ball = vball.GetDistance();
    double dir_diff = fabs(RadToDeg(vball.GetAngle()));
    double factor = cfg->kick_power_rate * (1.0 - 0.25 * dir_diff / 180.0 - 0.25 * (dist_ball - cfg->player_size - cfg->ball_size) / cfg->kickable_margin);

    C2DVector vkick;
    vkick.SetAngleDistance(dir, accel * factor);
//    vkick -= pBall->GetSpeed();
    double power2 = vkick.GetDistance() / factor;
    double dir2 = vkick.GetAngle();

    if (power2 > 100) power2 = 100;

    Kick(pa, power2, dir2);

}




void CBasicSkills::TurnTo(CLocalAgent *pa, C2DVector pos)
{
    C2DVector topos = Rot(pos - (C2DVector)pa->GetPosition(), -pa->GetDirection());
    Turn2(pa, topos.GetAngle());
}


void CBasicSkills::TurnNeckTo(CLocalAgent *pa, C2DVector pos)
{
    C2DVector necktopos = Rot(pos - (C2DVector)pa->GetPosition(), pa->GetNeckDirection());
    TurnNeck(pa, necktopos.GetAngle());
}

void CBasicSkills::TurnNeckToBall(CLocalAgent *pa)
{
    CBall *pBall = pa->GetWorldModel()->GetBall();
    CDiffuse2DVector vball = pBall->GetPosition() - pa->GetPosition();
    double necktoball = vball.GetAngle() - pa->GetNeckDirection();
    TurnNeck(pa, necktoball);
}

void CBasicSkills::TurnNeckStraight(CLocalAgent *pa)
{
    double neckang =  pa->GetDirection() - pa->GetNeckDirection();
    if ((neckang < DegToRad(-1.0)) || (neckang > DegToRad(1.0))) TurnNeck(pa, neckang);
}


void CBasicSkills::GoTo(CLocalAgent *pa, C2DVector pos, double accel)
{
    C2DVector topos = Rot(pos - (C2DVector)pa->GetPosition(), -pa->GetDirection());
    double ang = topos.GetAngle();

    // zistujeme, ci je v ceste nejaky hrac
    CAgent *paiw = NULL;
//    CAgent *paiw = PlayerInWay(pa, ang);
    if (paiw) {
        // tu sa riesi vyhybanie sa hracovi
        C2DVector toplr = Rot(paiw->GetPosition() - pa->GetPosition(), -pa->GetDirection());
        double pldist = toplr.GetDistance();
        double pldir = toplr.GetAngle();
        fprintf(m_ff, "player in way - dist %lf, dir %lf\n",pldist,pldir);

        // vyhybame sa iba ked je hrac blizsie ako miesto kam chceme ist
        if (pldist < topos.GetDistance())
            (pldir > 0) ? ang -= (MinimalDirection(pldist) - pldir)
                        : ang += (MinimalDirection(pldist) + pldir);
    }

    double mindir = DegToRad(5.0); // aby sme sa stale netocili
    if ((ang > -mindir) && (ang < mindir)) 
    {
        if (topos.GetDistance() > cfg->kickable_area) 
        {
            if (accel==0) {
                // vypocet accel, aby tam agent presne zastavil
                accel = (topos - (C2DVector)pa->GetSpeed()).GetDistance();
            }
            Dash2(pa, accel);
//            Dash(pa, spd);
        } 
        else 
        {
            LOGMSG(1, "ERROR Skill 'GoTo': already there.\n");
            return;
        }
    }
    else 
    {
        Turn2(pa, ang);
    }
}

void CBasicSkills::KickTo(CLocalAgent *pa, C2DVector pos, double accel)
{
/*
    C2DVector topos = Rot(pos - (C2DVector)pa->GetPosition(), -pa->GetDirection());
//    Kick2(pa, pow, topos.GetAngle());
    Kick(pa, pow, topos.GetAngle());
*/
    CBall *pBall = pa->GetWorldModel()->GetBall();
    C2DVector toball = Rot(pBall->GetPosition() - pa->GetPosition(), -pa->GetDirection());

    C2DVector topos = Rot(pos - (C2DVector)pa->GetPosition(), -pa->GetDirection());
    double ang = (topos - toball).GetAngle();
    if (accel==0) {
        // vypocet accel zo vzdialenosti pos aby tam lopta akurat prisla
        // accel = |vektor_k_pos - vektor_k_lopte|
        accel = (topos - toball).GetDistance();
    }
//    Kick2(pa, accel, ang);
    Kick2(pa, accel, ang);
}


void CBasicSkills::DribbleTo(CLocalAgent *pa, C2DVector pos, double spd, double safe)
{
    // Dorobit vyhybanie sa ostatnym hracom!

    if (BallKickable(pa)) {
//        TurnNeckTo(pa, pos);
        KickTo(pa, pos, spd);

    } else {
//        TurnNeckStraight(pa);
        C2DVector ballpos = pa->GetWorldModel()->GetBall()->GetPosition();
        double pow = spd;   //taka sila, aby zastal tesne pred loptou
        GoTo(pa, ballpos, pow);
    }

}



bool CBasicSkills::BallKickable(CLocalAgent *pa)
{
    CBall *pBall = pa->GetWorldModel()->GetBall();
    CDiffuse2DVector vball = Rot(pBall->GetPosition() - pa->GetPosition(), -pa->GetDirection());
    double disttoball = vball.GetDistance();
    double diffuball = vball.GetDiffusion();
    double dist = cfg->kickable_area - disttoball;
    /*
    if (dist < -diffuball) return 0;
    if (dist >= diffuball) return 1;
    return fabs(dist)/2.0;
    */
    if (dist < 0) return false;
    return true;
}

//! vzdialenost, 
double KickDistance(CLocalAgent *pa)
{

    return 20;
}

bool CBasicSkills::NearGoal(CLocalAgent *pa, EGoalType g)
{
    /*
    C2DVector p = pa->GetPosition();
    C2DVector k;
    k.SetAngleDistance(pa->GetDirection(), 0.90*KickDistance(pa));



    switch(g) {
    case GT_L:
        break;
    case GT_R:
        break;
    default:
    }



    CDiffuse2DVector goal(goalpos, 0);
    CDiffuse2DVector vgoal = goal - pa->GetPosition();
    double disttogoal = vgoal.GetDistance();
    double diffugoal = vgoal.GetDiffusion();
    double dist = 20 - disttogoal; // FIXME - nie 20, ale nejako vyratat odkial uz mozem strielat do brany
    // TODO nepocitat s brankou ako s bodom ale s celou brankou!
    if (dist < 0) return false;
    */
    return true;
}


//! Vrati minimalny uhol, v ktorom moze byt relativne ku mne iny hrac aby som donho nenarazil
double CBasicSkills::MinimalDirection(double dist)
{
    return asin(cfg->player_size / (dist + cfg->player_size));
}


CAgent *CBasicSkills::PlayerInWay(CLocalAgent *pa, double dir)
{
    CWorldModel *pwm = pa->GetWorldModel();
    Agents *pia = pwm->GetIdentifiedAgents();
    Agents *pua = pwm->GetUnidentifiedAgents();

    int i;
    C2DVector toplr, toplr2;
    double pldir;
    double pldist, pldist2;
    double mindir;
    CAgent *pnow = NULL;
    CAgent *pret = NULL;

    fprintf(m_ff, "search 4 player in way (see total %d players)\n", pia->size()+pua->size());


    // hladame medzi identifikovanymi hracmi
	for(i=0; i < pia->size(); i++)
	{
        pnow = pia->at(i);
        // vektor na hraca
        toplr = Rot(pnow->GetPosition() - pa->GetPosition(), -pa->GetDirection());
        pldir = toplr.GetAngle();
        pldist = toplr.GetDistance();
        mindir = MinimalDirection(pldist);
        if ((pldir >= -mindir) && (pldir <= mindir)) {
            if (pret) {
                toplr2 = Rot(pret->GetPosition() - pa->GetPosition(), -pa->GetDirection());
                pldist2 = toplr2.GetDistance();
                // vratime toho co je blizsie
                if( pldist < pldist2) pret = pnow;
            }
        }
    }

    // hladame medzi neidentifikovanymi hracmi
    for(i=0;i<pua->size();i++)
	{
        pnow = pua->at(i);
        // vektor na hraca
        toplr = Rot(pnow->GetPosition() - pa->GetPosition(), -pa->GetDirection());
        pldir = toplr.GetAngle();
        pldist = toplr.GetDistance();
        mindir = MinimalDirection(pldist);
        if ((pldir >= -mindir) && (pldir <= mindir)) {
            if (pret) {
                toplr2 = Rot(pret->GetPosition() - pa->GetPosition(), -pa->GetDirection());
                pldist2 = toplr2.GetDistance();
                // vratime toho co je blizsie
                if( pldist < pldist2) pret = pnow;
            }
        }
    }

    return pret;
}

/*
double CBasicSkills::NearestOpponentToBall(CLocalAgent *pa)
{


    return 0;
}
*/


